/*
 * Tool name   : FFLogDump
 * Description : A tool to extract history, cookie and favorites
 *               data of firefox 3.* logfile.
 * Author      : carrumba
 * Webpage     : http://www.megapanzer.com
 * Version     : 0.1
 * OS          : Tested on Microsoft Windows XP
 * Todo        : 
 *
 * Changes     : -
 * 
 *
 *
 *
 * License     :
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <windows.h>
#include <shlwapi.h>
#include <stdio.h>
#include "sqlite3.h"

#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>



#define snprintf _snprintf
#define MAX_BUF_SIZE 1024
#define FIREFOX_APP_DIR "\\Mozilla\\Firefox\\Profiles\\"
#define MAX_ARR_SIZE 128
#define FF_COOKIES_DB "cookies.sqlite"
#define FF_HISTORY_DB "places.sqlite"
#define FF_BOOKMARKS_DB "places.sqlite"
#define FF_DOWNLOADS_DB "downloads.sqlite"



/*
 * Constants, data types and function forward declarations.
 *
 */

#pragma comment(lib, "shlwapi.lib")


void printUsage(char *pProgName);
int getFFDefaultUserDir(char *pUserDir, int pUserDirSize);
void formatTimeStamp(char *pTimeStamp, char *pTSBuffer, int pTSBufferLength);
static int callbackFFCookies(void *NotUsed, int argc, char **argv, char **azColName);
static int callbackFFHistory(void *NotUsed, int argc, char **argv, char **azColName);
static int callbackFFBookmarks(void *NotUsed, int argc, char **argv, char **azColName);
static int callbackFFDownloads(void *NotUsed, int argc, char **argv, char **azColName);



/*
 * Program entry point
 *
 */


int main(int argc, char* argv[])
{
  int lRetVal = 0;
  sqlite3 *lDB;
  char *lErrMsg = 0;
  char lFFUserDir[MAX_BUF_SIZE + 1];
  char lDBFileFullPath[MAX_BUF_SIZE + 1];



  /*
   * Check command line parameters.
   *
   */

  if (argc < 2)
  {
    printUsage(argv[0]);
    lRetVal = 1;
    goto END;
  } else if (strncmp(argv[1], "-hi", 3) != 0 && 
             strncmp(argv[1], "-co", 3) != 0 &&
             strncmp(argv[1], "-bm", 3) != 0 &&
             strncmp(argv[1], "-dl", 3) != 0 &&
             strncmp(argv[1], "-h", 2) != 0 &&
             strncmp(argv[1], "-all", 4) != 0) {
    printUsage(argv[0]);
    lRetVal = 2;
    goto END;
  }



  /*
   * Find FF default user dir and extract requested log data.
   *
   */

  ZeroMemory(lFFUserDir, sizeof(lFFUserDir));
  if (getFFDefaultUserDir(lFFUserDir, sizeof(lFFUserDir)) == 0)
  {

    /*
     * List all Firefox cookies.
     *
     */

    if (strncmp(argv[1], "-co", 3) == 0 || strncmp(argv[1], "-all", 4) == 0)
	{
      printf("\r\nFirefox cookies\r\n_______________\r\n\r\n");
      ZeroMemory(lDBFileFullPath, sizeof(lDBFileFullPath));
      snprintf(lDBFileFullPath, sizeof(lDBFileFullPath) - 1, "%s%s", lFFUserDir, FF_COOKIES_DB);
	  printf("-co : drinn 1 : %s\n", lDBFileFullPath);

      if(sqlite3_open(lDBFileFullPath, &lDB) == SQLITE_OK)
      {
        if(sqlite3_exec(lDB, "SELECT name,value,host,path,expiry FROM moz_cookies", callbackFFCookies, 0, &lErrMsg))
           sqlite3_free(lErrMsg); 

        sqlite3_close(lDB);
	  } else {
        lRetVal = 3;
        goto END;
	  } // if(sqlite3_open...
	} // if (strncmp(argv[1]..



    /*
     * List Firefox download history.
     *
     */

    if (strncmp(argv[1], "-dl", 3) == 0 || strncmp(argv[1], "-all", 4) == 0)
	{
      printf("\r\nFirefox download history\r\n________________________\r\n\r\n");
      ZeroMemory(lDBFileFullPath, sizeof(lDBFileFullPath));
      snprintf(lDBFileFullPath, sizeof(lDBFileFullPath) - 1, "%s%s", lFFUserDir, FF_DOWNLOADS_DB);
	  printf("-co : drinn 1 : %s\n", lDBFileFullPath);

      if(sqlite3_open(lDBFileFullPath, &lDB) == SQLITE_OK)
      {
        if(sqlite3_exec(lDB, "SELECT name,source,target,mimeType,maxBytes FROM moz_downloads", callbackFFDownloads, 0, &lErrMsg))
           sqlite3_free(lErrMsg); 

        sqlite3_close(lDB);
	  } else {
        lRetVal = 3;
        goto END;
	  } // if(sqlite3_open...
	} // if (strncmp(argv[1]..





    /*
     * List Firefox history.
     *
     */

    if (strncmp(argv[1], "-hi", 3) == 0 || strncmp(argv[1], "-all", 4) == 0)
	{
      printf("\r\n\r\n\r\nFirefox history\r\n_______________\r\n\r\n");
      ZeroMemory(lDBFileFullPath, sizeof(lDBFileFullPath));
      snprintf(lDBFileFullPath, sizeof(lDBFileFullPath) - 1, "%s%s", lFFUserDir, FF_HISTORY_DB);

      if(sqlite3_open(lDBFileFullPath, &lDB) == SQLITE_OK)
      {
        if(sqlite3_exec(lDB, "SELECT title,url,last_visit_date FROM moz_places", callbackFFHistory, 0, &lErrMsg))
           sqlite3_free(lErrMsg); 

        sqlite3_close(lDB);
	  } else {
        lRetVal = 4;
        goto END;
	  } // if(sqlite3_open...
	} // if (strncmp(argv[1]..






    /*
     * List Firefox bookmarks.
     *
     */

    if (strncmp(argv[1], "-bm", 3) == 0 || strncmp(argv[1], "-all", 4) == 0)
	{
      printf("\r\n\r\n\r\nFirefox bookmarks\r\n_________________\r\n\r\n");
      ZeroMemory(lDBFileFullPath, sizeof(lDBFileFullPath));
      snprintf(lDBFileFullPath, sizeof(lDBFileFullPath) - 1, "%s%s", lFFUserDir, FF_BOOKMARKS_DB);

      if(sqlite3_open(lDBFileFullPath, &lDB) == SQLITE_OK)
      {
        if(sqlite3_exec(lDB, "SELECT BM.title, PC.URL FROM moz_bookmarks BM, moz_places PC  where BM.fk = PC.ID", callbackFFBookmarks, 0, &lErrMsg))
           sqlite3_free(lErrMsg); 

        sqlite3_close(lDB);
 	  } else {
        lRetVal = 5;
        goto END;
	  } // if(sqlite3_open...
	} // if (strncmp(argv[1]..

  } else {
    printf("Error occurred while determining Firefox default user directory!\n");
    lRetVal = 2;
    goto END;
  } // if (getFFDe...


END:

  return(lRetVal);
}




/*
 * Tool usage.
 *
 */

void printUsage(char *pProgName)
{
  printf("Usage  : %s -all|-bm|-co|-h|-hi\n\n\n\n", pProgName);
  printf("Examples ...\n\n\n");
  printf("%s -bm  : print bookmarks\n", pProgName);
  printf("%s -h  : print help\n", pProgName);
  printf("%s -hi  : print browser history\n", pProgName);
  printf("%s -co  : print cookies\n", pProgName);
  printf("%s -dl  : print download history\n", pProgName);
  printf("%s -all : print all logs\n\n", pProgName);
}




/*
 * Determine the Firefox directory.
 *
 */

int getFFDefaultUserDir(char *pUserDir, int pUserDirSize)
{
  int lRetVal = 0;
  char lBaseDir[MAX_BUF_SIZE + 1];
  char lLogFileBaseDir[MAX_BUF_SIZE + 1];
  char lDirPattern[MAX_BUF_SIZE + 1];
  HANDLE hFind = INVALID_HANDLE_VALUE;
  WIN32_FIND_DATA lFindFileData;


  /*
   * Initialisation stuff.
   *
   */

  ZeroMemory(lBaseDir, sizeof(lBaseDir));
  ZeroMemory(lDirPattern, sizeof(lDirPattern));

  snprintf(lBaseDir, sizeof(lBaseDir) - 1, "%s\\%s\\", getenv("APPDATA"), FIREFOX_APP_DIR);
  snprintf(lDirPattern, sizeof(lDirPattern) - 1, "%s\\%s\\*", getenv("APPDATA"), FIREFOX_APP_DIR);



  if ((hFind = FindFirstFile(lDirPattern, &lFindFileData)) == INVALID_HANDLE_VALUE) 
  {
    lRetVal = 1;
    goto END;
  } else {
    do 
    {
	  if((lFindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) &&
		  strncmp(lFindFileData.cFileName, "..", 2) &&
		  strncmp(lFindFileData.cFileName, ".", 1) &&
          strstr(lFindFileData.cFileName, ".default"))
	  {
        ZeroMemory(lLogFileBaseDir, sizeof(lLogFileBaseDir));
        snprintf(lLogFileBaseDir, sizeof(lLogFileBaseDir) - 1, "%s%s\\", lBaseDir, lFindFileData.cFileName);
        strncpy(pUserDir, lLogFileBaseDir, pUserDirSize);
        break;
      } // if(lFindFileData.dwFi...
    } while(FindNextFile(hFind, &lFindFileData) != 0);

    FindClose(hFind);
  }

END:



  return(lRetVal);
}




/*
 * Callback function to list all Firefox downloads.
 *
 */

static int callbackFFDownloads(void *NotUsed, int argc, char **argv, char **azColName)
{
  int lCounter = 0;


  for(; lCounter < argc; lCounter += 5)
  {
    if (argv[lCounter])
	{
      printf("Name      : %s\n", argv[lCounter]?argv[lCounter]:"");
      printf("Source    : %s\n", argv[lCounter+1]?argv[lCounter+1]:"");
      printf("Target    : %s\n", argv[lCounter+2]?argv[lCounter+2]:"");
      printf("Mime type : %s\n", argv[lCounter+3]?argv[lCounter+3]:"");
      printf("Size      : %s\n\n", argv[lCounter+4]?argv[lCounter+4]:"");
	} // if (argv[lCounter]...
  } // for(; lCoun...


  return(0);
}



/*
 * Callback function to list all Firefox cookie entries.
 *
 */

static int callbackFFCookies(void *NotUsed, int argc, char **argv, char **azColName)
{
  int lCounter = 0;


  for(; lCounter < argc; lCounter += 5)
  {
    if (argv[lCounter])
	{
      printf("Name    : %s\n", argv[lCounter]?argv[lCounter]:"");
      printf("Value   : %s\n", argv[lCounter+1]?argv[lCounter+1]:"");
      printf("Host    : %s\n", argv[lCounter+2]?argv[lCounter+2]:"");
      printf("Path    : %s\n", argv[lCounter+3]?argv[lCounter+3]:"");
      printf("Expiry  : %s\n\n", argv[lCounter+4]?argv[lCounter+4]:"");
	} // if (argv[lCounter]...
  } // for(; lCoun...


  return(0);
}





/*
 * Callback function to list all Firefox cookie entries. 
 *
 */

static int callbackFFHistory(void *NotUsed, int argc, char **argv, char **azColName)
{
  int lCounter = 0;
  char lTemp[MAX_BUF_SIZE + 1];
  char lTimeStamp[MAX_BUF_SIZE + 1];

  for(; lCounter < argc; lCounter += 3)
  {
    if (argv[lCounter+1])
	{
      /*
       * Generate human readable timestamp.
       *
       */

      ZeroMemory(lTimeStamp, sizeof(lTimeStamp));
      ZeroMemory(lTemp, sizeof(lTemp));

      if (argv[lCounter+2])
	  {
        CopyMemory(lTemp, argv[lCounter+2], 10);
        formatTimeStamp(lTemp, lTimeStamp, sizeof(lTimeStamp));
	  }

      printf("Title        : %s\n", argv[lCounter]?argv[lCounter]:"");
      printf("URL          : %s\n", argv[lCounter+1]?argv[lCounter+1]:"");
      printf("Last visited : %s\n\n", lTimeStamp);
	} // if (argv[lCounter]...
  } // for(; lCoun...


  return(0);
}






/*
 * Callback function to list all Firefox bookmark/favorites entries. 
 *
 */

static int callbackFFBookmarks(void *NotUsed, int argc, char **argv, char **azColName)
{
  int lCounter = 0;


  for(; lCounter < argc; lCounter += 2)
  {
    if (argv[lCounter])
	{
      printf("Title   : %s\n", argv[lCounter]?argv[lCounter]:"");
      printf("URL     : %s\n\n", argv[lCounter+1]?argv[lCounter+1]:"");
	} // if (argv[lCounter]...
  } // for(; lCoun...


  return(0);
}





/*
 * Transform unix time stamp to human readable time stamp.
 *
 */

void formatTimeStamp(char *pTimeStamp, char *pTSBuffer, int pTSBufferLength)
{
  time_t lTime = 0;
  struct tm *lTS;

  lTime = atoi(pTimeStamp);
  lTS = localtime(&lTime);
  strftime(pTSBuffer, pTSBufferLength, "%Y.%m.%d %H:%M:%S", lTS);
}